__author__ = 'Robert Cope', 'Jochen Hoenicke'

import os
import sys
import time
import usb1
import array
import libusb1
import threading
from struct import pack

from hantekosc.py_ht6022.Firmware import dso6021_firmware,dso6022be_firmware, dso6022bl_firmware, fx2_ihex_to_control_packets

class Oscilloscope(object):
    # sync with "__version__" in "setup.py" and "FIRMWARE_VERSION" in "py_ht6022/Firmware/DSO6022BE/descriptor.inc"
    FIRMWARE_VERSION = 0x0210
    NO_FIRMWARE_VENDOR_ID = 0x04B4
    FIRMWARE_PRESENT_VENDOR_ID = 0x04B5
    PRODUCT_ID_21 = 0x6021
    PRODUCT_ID_BE = 0x6022
    PRODUCT_ID_BL = 0x602A

    RW_FIRMWARE_REQUEST = 0xa0
    RW_FIRMWARE_INDEX = 0x00

    RW_EEPROM_REQUEST = 0xa2
    RW_EEPROM_INDEX = 0x00

    SET_SAMPLE_RATE_REQUEST = 0xe2
    SET_SAMPLE_RATE_VALUE = 0x00
    SET_SAMPLE_RATE_INDEX = 0x00

    SET_CH1_VR_REQUEST = 0xe0
    SET_CH1_VR_VALUE = 0x00
    SET_CH1_VR_INDEX = 0x00

    SET_CH2_VR_REQUEST = 0xe1
    SET_CH2_VR_VALUE = 0x00
    SET_CH2_VR_INDEX = 0x00

    TRIGGER_REQUEST = 0xe3
    TRIGGER_VALUE = 0x00
    TRIGGER_INDEX = 0x00

    SET_NUMCH_REQUEST = 0xe4
    SET_NUMCH_VALUE = 0x00
    SET_NUMCH_INDEX = 0x00

    SET_AC_DC_REQUEST = 0xe5
    SET_AC_DC_VALUE = 0x00
    SET_AC_DC_INDEX = 0x00

    AC = 0
    DC = 1
    AC_AC = 0x00
    AC_DC = 0x01
    DC_AC = 0x10
    DC_DC = 0x11

    SET_CAL_FREQ_REQUEST = 0xe6
    SET_CAL_FREQ_VALUE = 0x00
    SET_CAL_FREQ_INDEX = 0x00

    CALIBRATION_EEPROM_OFFSET   = 0x08
    CALIBRATION_EEPROM_SIZE     = 32
    CALIBRATION_EEPROM_EXT_SIZE = 80


    SAMPLE_RATES = {
                    102: ( "20 kS/s",  20e3),
                    103: ( "32 kS/s",  32e3),
                    104: ( "40 kS/s",  40e3),
                    105: ( "50 kS/s",  50e3),
                    106: ( "64 kS/s",  64e3),
                    110: ("100 kS/s", 100e3),
                    113: ("128 kS/s", 128e3),
                    120: ("200 kS/s", 200e3),
                    140: ("400 kS/s", 400e3),
                    150: ("500 kS/s", 500e3),
                    164: ("640 kS/s", 640e3),
                      1: (  "1 MS/s",   1e6),
                      2: (  "2 MS/s",   2e6),
                      3: (  "3 MS/s",   3e6),
                      4: (  "4 MS/s",   4e6),
                      5: (  "5 MS/s",   5e6),
                      6: (  "6 MS/s",   6e6),
                      8: (  "8 MS/s",   8e6),
                     10: ( "10 MS/s",  10e6),
                     12: ( "12 MS/s",  12e6),
                     15: ( "15 MS/s",  15e6),
                     16: ( "16 MS/s",  16e6),
                     24: ( "24 MS/s",  24e6),
                     30: ( "30 MS/s",  30e6),
                     48: ( "48 MS/s",  48e6)
                    }

    VOLTAGE_RANGES = { 1: ('+/- 5V', 0.0390625, 2.5),
                       2: ('+/- 2.5V', 0.01953125, 1.25),
                       5: ('+/- 1V', 0.0078125, 0.5),
                      10: ('+/- 500mV', 0.00390625, 0.25)
                     }

#    CAL_FREQUENCYS = {
#                      105: ( " 50 Hz",     50 ),
#                      106: ( " 60 Hz",     60 ),
#                      110: ( "100 Hz",    100 ),
#                      120: ( "200 Hz",    200 ),
#                      150: ( "500 Hz",    500 ),
#                        1: (  "1 kHz",   1000 ),
#                        2: (  "2 kHz",   2000 ),
#                        5: (  "5 kHz",   5000 ),
#                       10: ( "10 kHz",  10000 ),
#                       20: ( "20 kHz",  20000 ),
#                       50: ( "50 kHz",  50000 ),
#                      100: ("100 kHz", 100000 )
#                     }


    # defaults to 6022BE with the possibility to supply a non standard VID/PID combination
    def __init__(self, VID=NO_FIRMWARE_VENDOR_ID, PID=PRODUCT_ID_BE):
        self.device = None
        self.device_handle = None
        self.context = usb1.USBContext()
        self.is_device_firmware_present = False
        self.supports_single_channel = False
        self.is_iso = False
        self.packetsize = None
        self.num_channels = 2
        self.ac_dc_status = 0x11
        self.VID=VID
        self.PID=PID
        self.calibration = None
        self.calibration_ext = None
        self.offset1 = { 1:0, 2:0, 5:0, 10:0 }
        self.offset2 = { 1:0, 2:0, 5:0, 10:0 }
        self.gain1 = { 1:1.01, 2:1.01, 5:0.99, 10:1.0 }
        self.gain2 = { 1:1.01, 2:1.01, 5:0.99, 10:1.0 }


    def setup(self):
        """
        Attempt to find a suitable scope to run.
        :return: True if a {6022BE, 6022BL, 6021} (or user defined) scope was found, False otherwise.
        """
        # look for a user defined device that doesn't match {6021,6022BE,6022BL}
        # ToDo: can't create a device list if always the first matching device is used
        # ToDo: rather initialize self.VID to None?
        if ( ( self.VID != self.NO_FIRMWARE_VENDOR_ID and self.VID != self.FIRMWARE_PRESENT_VENDOR_ID )
        or ( self.PID != self.PRODUCT_ID_21
         and self.PID != self.PRODUCT_ID_BE
         and self.PID != self.PRODUCT_ID_BL ) ):
            self.device = (
                self.context.getByVendorIDAndProductID(
                    self.VID, self.PID, skip_on_error=True, skip_on_access_error=True
                )
            )
            if self.device:
                self.is_device_firmware_present = False
                return True

        # 1st look for 6022BE
        self.device = (
            self.context.getByVendorIDAndProductID(
                self.FIRMWARE_PRESENT_VENDOR_ID, self.PRODUCT_ID_BE, skip_on_error=True, skip_on_access_error=True
            )
         or self.context.getByVendorIDAndProductID(
                self.NO_FIRMWARE_VENDOR_ID, self.PRODUCT_ID_BE, skip_on_error=True, skip_on_access_error=True
            )
        )
        if self.device:
            self.supports_single_channel = \
            self.is_device_firmware_present = (
                  self.device.getVendorID() == self.FIRMWARE_PRESENT_VENDOR_ID
              and self.device.getbcdDevice() == self.FIRMWARE_VERSION
            ) # latest custom FW loaded
            return True

        # if not found look for 6022BL
        self.device = (
            self.context.getByVendorIDAndProductID(
                self.FIRMWARE_PRESENT_VENDOR_ID, self.PRODUCT_ID_BL, skip_on_error=True, skip_on_access_error=True
            )
         or self.context.getByVendorIDAndProductID(
                self.NO_FIRMWARE_VENDOR_ID, self.PRODUCT_ID_BL, skip_on_error=True, skip_on_access_error=True
            )
        )
        if self.device:
            self.supports_single_channel = \
            self.is_device_firmware_present = (
                  self.device.getVendorID() == self.FIRMWARE_PRESENT_VENDOR_ID
              and self.device.getbcdDevice() == self.FIRMWARE_VERSION
            ) # latest custom FW loaded
            return True

        # if not found look for 6021
        self.device = (
            self.context.getByVendorIDAndProductID(
                self.FIRMWARE_PRESENT_VENDOR_ID, self.PRODUCT_ID_21, skip_on_error=True, skip_on_access_error=True
            )
         or self.context.getByVendorIDAndProductID(
                self.NO_FIRMWARE_VENDOR_ID, self.PRODUCT_ID_21, skip_on_error=True, skip_on_access_error=True
            )
        )
        if self.device:
            self.supports_single_channel = \
            self.is_device_firmware_present = (
                  self.device.getVendorID() == self.FIRMWARE_PRESENT_VENDOR_ID
              and self.device.getbcdDevice() == self.FIRMWARE_VERSION
            ) # latest custom FW loaded
            return True

        return False


    def open_handle(self):
        """
        Open a device handle for the scope. This needs to occur before sending any commands.
        :return: True if successful, False otherwise. May raise various libusb exceptions on fault.
        """
        if self.device_handle:
            return True
        if not self.device and not self.setup():
            return False
        self.device_handle = self.device.open()
        if os.name == 'posix' and self.device_handle.kernelDriverActive(0):
            self.device_handle.detachKernelDriver(0)
        try:
            self.device_handle.claimInterface(0)
        except Exception as e:
            print( e, file=sys.stderr )
            return False
        if self.is_device_firmware_present:
            self.set_num_channels(2)
            self.set_interface(0)
        return True


    def close_handle(self, release_interface=True):
        """
        Close the current scope device handle. This should always be called at clean-up.
        :param release_interface: (OPTIONAL) Attempt to release the interface, if we still have it.
        :return: True if successful. May assert or raise various libusb errors if something went wrong.
        """
        if not self.device_handle:
            return True
        try:
            if release_interface:
                self.device_handle.releaseInterface(0)
            self.device_handle.close()
        except Exception as e:
            print( e, file=sys.stderr )
            return False
        self.device_handle = None
        return True


    def get_fw_version( self ):
        """
        Returns the BCD coded firmware/device version
        """
        if not self.device:
            return None
        return self.device.getbcdDevice()


    def get_product_string( self ):
        """
        Returns the product name string
        """
        if not self.device:
            return None
        return self.device.getProduct()


    def get_serial_number_string( self ):
        """
        Returns the 6 byte serial number as 12 char string
        """
        if not self.device:
            return None
        return self.device.getSerialNumber()


    def __del__(self):
        self.close_handle()


    def poll(self):
        self.context.handleEvents()


    def flash_firmware(self, firmware=None, supports_single_channel=True, timeout=60):
        """
        Flash scope firmware to the target scope device. This needs to occur once when the device is first attached,
        as the 6022BE does not have any persistant storage.
        :param firmware: (OPTIONAL) The firmware packets to send. Default: custom firmware (either BE or BL).
        :param supports_single_channel: (OPTIONAL) Set this to false if loading the stock firmware, as it does not
                                        support reducing the number of active channels.
        :param timeout: (OPTIONAL) A timeout for each packet transfer on the firmware upload. Default: 60 seconds.
        :return: True if the correct device vendor was found after flashing firmware, False if the default Vendor ID
                 was present for the device. May assert or raise various libusb errors if something went wrong.
        """
        if not self.device_handle:
            assert self.open_handle()
        if not firmware: # called without an explicit firmware parameter
            if self.device.getProductID() == self.PRODUCT_ID_BE:
                firmware = dso6022be_firmware
            elif self.device.getProductID() == self.PRODUCT_ID_BL:
                firmware = dso6022bl_firmware
            elif self.device.getProductID() == self.PRODUCT_ID_21:
                firmware = dso6021_firmware
            else:
                return False
        for packet in firmware:
            bytes_written = self.device_handle.controlWrite(0x40, self.RW_FIRMWARE_REQUEST,
                                                            packet.value, self.RW_FIRMWARE_INDEX,
                                                            packet.data, timeout=timeout)
            assert bytes_written == packet.size
        # After firmware is written, scope will typically show up again as a different device, so scan again
        self.close_handle(release_interface=False)
        time.sleep(.5)
        while not self.setup():
            time.sleep(0.1)
        self.supports_single_channel = supports_single_channel
        self.open_handle()
        return self.is_device_firmware_present


    def flash_firmware_from_hex(self, hex_file, timeout=60):
        """
        Open an Intel hex file for the 8051 and try to flash it to the scope.
        :param hex_file: The hex file to load for flashing.
        :param timeout: (OPTIONAL) A timeout for each packet transfer on the firmware upload. Default: 60 seconds.
        :return: True if the correct device vendor was found after flashing firmware, False if the default Vendor ID
                 was present for the device. May assert or raise various libusb errors if something went wrong.
        """
        return self.flash_firmware(firmware=fx2_ihex_to_control_packets(hex_file), timeout=timeout)


    def read_firmware(self, address=0, length=8192, to_ihex=True, chunk_len=16, timeout=60):
        """
        Read the entire device RAM, and return a raw string.
        :param to_ihex: (OPTIONAL) Convert the firmware into the Intel hex format after reading. Otherwise, return
                        the firmware is a raw byte string. Default: True
        :param chunk_len: (OPTIONAL) The length of RAM chunks to pull from the device at a time. Default: 16 bytes.
        :param timeout: (OPTIONAL) A timeout for each packet transfer on the firmware upload. Default: 60 seconds.
        :return: The raw device firmware, if successful.
                 May assert or raise various libusb errors if something went wrong.
        """
        if not self.device_handle:
            assert self.open_handle()

        bytes_written = self.device_handle.controlWrite(0x40, self.RW_FIRMWARE_REQUEST,
                                                        0xe600, self.RW_FIRMWARE_INDEX,
                                                        b'\x01', timeout=timeout)
        assert bytes_written == 1
        firmware_chunk_list = []
        # TODO: Fix this for when 8192 isn't divisible by chunk_len
        for packet in range(0, length//chunk_len):
            chunk = self.device_handle.controlRead(0x40, self.RW_FIRMWARE_REQUEST,
                                                   address + packet * chunk_len, self.RW_FIRMWARE_INDEX,
                                                   chunk_len, timeout=timeout)
            firmware_chunk_list.append(chunk)
            assert len(chunk) == chunk_len
        bytes_written = self.device_handle.controlWrite(0x40, self.RW_FIRMWARE_REQUEST,
                                                        0xe600, self.RW_FIRMWARE_INDEX,
                                                        b'\x00', timeout=timeout)
        assert bytes_written == 1
        if not to_ihex:
            return ''.join(firmware_chunk_list)

        lines = []
        for i, chunk in enumerate(firmware_chunk_list):
            addr = address + i*chunk_len
            iterable_chunk = array.array('B', chunk)
            hex_data = "".join(["{:02x}".format(b) for b in iterable_chunk])
            total_sum = (sum(iterable_chunk) + chunk_len + (addr % 256) + (addr >> 8)) % 256
            checksum = (((0xFF ^ total_sum) & 0xFF) + 0x01) % 256
            line = ":{nbytes:02x}{addr:04x}{itype:02x}{hex_data}{checksum:02x}".format(nbytes=chunk_len,
                                                                                        addr=addr,
                                                                                        itype=0x00,
                                                                                        hex_data=hex_data,
                                                                                        checksum=checksum)
            lines.append(line)
        # Add stop record at the end.
        lines.append(":00000001ff")
        return "\n".join(lines)


    def get_calibration_values(self, size=32, timeout=0):
        """
        Retrieve the current calibration values from the oscilloscope.
        Calculate the internal offset and gain correction used by "scale_read_data()"
        :param timeout: (OPTIONAL) A timeout for the transfer. Default: 0 (No timeout)
        :return: A size size single byte int list of calibration values, if successful.
                 May assert or raise various libusb errors if something went wrong.
        """
        # get the standard (factory) calibration values, these are always available
        self.calibration = array.array('B', self.read_eeprom(self.CALIBRATION_EEPROM_OFFSET, size, timeout=timeout))

        self.offset1 = {
          10: self.calibration[0]  - 128,
           5: self.calibration[6]  - 128,
           2: self.calibration[8]  - 128,
           1: self.calibration[14] - 128
        }

        self.offset2 = {
          10: self.calibration[1]  - 128,
           5: self.calibration[7]  - 128,
           2: self.calibration[9]  - 128,
           1: self.calibration[15] - 128
        }

        # get the extended calibration values that are set by the "calibration.py" program
        self.calibration_ext = array.array('B', self.read_eeprom(self.CALIBRATION_EEPROM_OFFSET, self.CALIBRATION_EEPROM_EXT_SIZE, timeout=timeout))

        # check for extended offset correction
        if self.calibration_ext[48] != 0 and self.calibration_ext[48] != 255:
                self.offset1[10] = self.offset1[10] +  ( self.calibration_ext[48] - 128 ) / 250
        if self.calibration_ext[54] != 0 and self.calibration_ext[54] != 255:
                self.offset1[5]  = self.offset1[5]  +  ( self.calibration_ext[54] - 128 ) / 250
        if self.calibration_ext[56] != 0 and self.calibration_ext[56] != 255:
                self.offset1[2]  = self.offset1[2]  +  ( self.calibration_ext[56] - 128 ) / 250
        if self.calibration_ext[62] != 0 and self.calibration_ext[62] != 255:
                self.offset1[1]  = self.offset1[1]  +  ( self.calibration_ext[62] - 128 ) / 250

        # check for gain correction
        if self.calibration_ext[32] != 0 and self.calibration_ext[32] != 255:
                self.gain1[10] = self.gain1[10] * ( 1 + ( self.calibration_ext[32] - 128 ) / 500 )
        if self.calibration_ext[38] != 0 and self.calibration_ext[38] != 255:
                self.gain1[5]  = self.gain1[5]  * ( 1 + ( self.calibration_ext[38] - 128 ) / 500 )
        if self.calibration_ext[40] != 0 and self.calibration_ext[40] != 255:
                self.gain1[2]  = self.gain1[2]  * ( 1 + ( self.calibration_ext[40] - 128 ) / 500 )
        if self.calibration_ext[46] != 0 and self.calibration_ext[46] != 255:
                self.gain1[1]  = self.gain1[1]  * ( 1 + ( self.calibration_ext[46] - 128 ) / 500 )

        # check for extended offset correction
        if self.calibration_ext[49] != 0 and self.calibration_ext[49] != 255:
                self.offset2[10] = self.offset2[10] +  ( self.calibration_ext[49] - 128 ) / 250
        if self.calibration_ext[55] != 0 and self.calibration_ext[55] != 255:
                self.offset2[5]  = self.offset2[5]  +  ( self.calibration_ext[55] - 128 ) / 250
        if self.calibration_ext[57] != 0 and self.calibration_ext[57] != 255:
                self.offset2[2]  = self.offset2[2]  +  ( self.calibration_ext[57] - 128 ) / 250
        if self.calibration_ext[63] != 0 and self.calibration_ext[63] != 255:
                self.offset2[1]  = self.offset2[1]  +  ( self.calibration_ext[63] - 128 ) / 250

        # check for gain correction
        if self.calibration_ext[33] != 0 and self.calibration_ext[33] != 255:
                self.gain2[10] = self.gain2[10] * ( 1 + ( self.calibration_ext[33] - 128 ) / 500 )
        if self.calibration_ext[39] != 0 and self.calibration_ext[39] != 255:
                self.gain2[5]  = self.gain2[5]  * ( 1 + ( self.calibration_ext[39] - 128 ) / 500 )
        if self.calibration_ext[41] != 0 and self.calibration_ext[41] != 255:
                self.gain2[2]  = self.gain2[2]  * ( 1 + ( self.calibration_ext[41] - 128 ) / 500 )
        if self.calibration_ext[47] != 0 and self.calibration_ext[47] != 255:
                self.gain2[1]  = self.gain2[1]  * ( 1 + ( self.calibration_ext[47] - 128 ) / 500 )

        return self.calibration


    def set_calibration_values(self, cal_list, timeout=0):
        """
        Set the a calibration level for the oscilloscope.
        :param cal_list: The list of calibration values, should usually be 32 single byte ints.
        :param timeout: (OPTIONAL) A timeout for the transfer. Default: 0 (No timeout)
        :return: True if successful. May assert or raise various libusb errors if something went wrong.
        """
        cal_list = cal_list if isinstance(cal_list, bytearray) else array.array('B', cal_list).tobytes()
        return self.write_eeprom(self.CALIBRATION_EEPROM_OFFSET, cal_list, timeout=timeout)


    def read_eeprom(self, offset, length, timeout=0):
        """
        Retrieve values from the eeprom on the oscilloscope.
        :param offset: start address into the eeprom.
        :param length: number of bytes to retrieve.
        :param timeout: (OPTIONAL) A timeout for the transfer. Default: 0 (No timeout)
        :return: The raw eeprom data.
        """
        if not self.device_handle:
            assert self.open_handle()
        data = self.device_handle.controlRead(0x40, self.RW_EEPROM_REQUEST, offset,
                                              self.RW_EEPROM_INDEX, length, timeout=timeout)
        return data


    def write_eeprom(self, offset, data, timeout=0):
        """
        Set the a calibration level for the oscilloscope.
        :param offset: The start address into the eeprom.
        :param data: The raw string of data to write.
        :param timeout: (OPTIONAL) A timeout for the transfer. Default: 0 (No timeout)
        :return: True if successful. May assert or raise various libusb errors if something went wrong.
        """
        if not self.device_handle:
            assert self.open_handle()
        data_len = self.device_handle.controlWrite(0x40, self.RW_EEPROM_REQUEST, offset,
                                                   self.RW_EEPROM_INDEX, data, timeout=timeout)
        assert data_len == len(data)
        return True


    def start_capture(self, timeout=0):
        """
        Tell the device to start capturing samples.
        :param timeout: (OPTIONAL) A timeout for the transfer. Default: 0 (No timeout)
        :return: True if successful. May assert or raise various libusb errors if something went wrong.
        """
        bytes_written = self.device_handle.controlWrite(0x40, self.TRIGGER_REQUEST,
                                                        self.TRIGGER_VALUE, self.TRIGGER_INDEX,
                                                        b'\x01', timeout=timeout)
        return bytes_written == 1


    def stop_capture(self, timeout=0):
        """
        Tell the device to stop capturing samples.  This is only supported
        with the custom firmware.
        :param timeout: (OPTIONAL) A timeout for the transfer. Default: 0 (No timeout)
        :return: True if successful. May assert or raise various libusb errors if something went wrong.
        """
        bytes_written = self.device_handle.controlWrite(0x40, self.TRIGGER_REQUEST,
                                                        self.TRIGGER_VALUE, self.TRIGGER_INDEX,
                                                        b'\x00', timeout=timeout)
        return bytes_written == 1


    def read_data(self, data_size=0x400, raw=False, timeout=0):
        """
        Read both channel's ADC data from the device. No trigger support, you need to do this in software.
        :param data_size: (OPTIONAL) The number of data points for each channel to retrieve. Default: 0x400 points.
        :param raw: (OPTIONAL) Return the raw bytestrings from the scope. Default: Off
        :param timeout: (OPTIONAL) The timeout for each bulk transfer from the scope. Default: 0 (No timeout)
        :return: If raw, two bytestrings are returned, the first for CH1, the second for CH2. If raw is off, two
                 lists are returned (by iterating over the bytestrings and converting to ordinals). The lists contain
                 the ADC value measured at that time, which should be between 0 - 255.

                 If you'd like nicely scaled data, just dump the return lists into the scale_read_data method with
                 your current voltage range setting.

                 This method may assert or raise various libusb errors if something went wrong.
        """
        data_size *= self.num_channels
        if not self.device_handle:
            assert self.open_handle()

        self.start_capture()
        data = self.device_handle.bulkRead(0x86, data_size, timeout=timeout)
        self.stop_capture()
        if self.num_channels == 2:
            chdata = data[::2], data[1::2]
        else:
            chdata = data, ''
        if raw:
            return chdata
        else:
            return array.array('B', chdata[0]), array.array('B', chdata[1])


    def build_data_reader(self, raw=False):
        """
        Build a (slightly) more optimized reader closure, for (slightly) better performance.
        :param raw: (OPTIONAL) Return the raw bytestrings from the scope. Default: Off
        :return: A fast_read_data function, which behaves much like the read_data function. The fast_read_data
                 function returned takes two parameters:
                 :param data_size: Number of data points to return (1 point <-> 1 byte).
                 :param timeout: (OPTIONAL) The timeout for each bulk transfer from the scope. Default: 0 (No timeout)
                 :return:  If raw, two bytestrings are returned, the first for CH1, the second for CH2. If raw is off,
                 two lists are returned (by iterating over the bytestrings and converting to ordinals).
                 The lists contain the ADC value measured at that time, which should be between 0 - 255.

        This method and the closure may assert or raise various libusb errors if something went/goes wrong.
        """
        if not self.device_handle:
            assert self.open_handle()
        scope_bulk_read = self.device_handle.bulkRead
        array_builder = array.array
        if self.num_channels == 1 and raw:
            def fast_read_data(data_size, timeout=0):
                data = scope_bulk_read(0x86, data_size, timeout)
                return data, ''
        elif self.num_channels == 1 and not raw:
            def fast_read_data(data_size, timeout=0):
                data = scope_bulk_read(0x86, data_size, timeout)
                return array_builder('B', data), array_builder('B', '')
        elif self.num_channels == 2 and raw:
            def fast_read_data(data_size, timeout=0):
                data_size <<= 0x1
                data = scope_bulk_read(0x86, data_size, timeout)
                return data[::2], data[1::2]
        elif self.num_channels == 2 and not raw:
            def fast_read_data(data_size, timeout=0):
                data_size <<= 0x1
                data = scope_bulk_read(0x86, data_size, timeout)
                return array_builder('B', data[::2]), array_builder('B', data[1::2])
        else:
            # Should never be here.
            assert False
        return fast_read_data


    def set_interface(self, alt):
        """
        Set the alternative interface (bulk or iso) to use.  This is only
        supported with the custom firmware.  Iso stands for isochronous usb
        transfer and this guarantees a bandwidth and therefore make gapless
        sampling more likely.
        :param int alt: The interface to use.  Values are
                           0 <-> Bulk transfer
                           1 <-> Fastest Iso transfer
                           2 <-> Slower Iso transfer (needs less usb bandwidth)
                           3 <-> Even slower Iso transfer
        :return: True if successful. May assert or raise various libusb errors if something went wrong.
        """
        if not self.device_handle:
            assert self.open_handle()
        self.device_handle.setInterfaceAltSetting(0, alt)
        endpoint_info = self.device[0][0][alt][0]
        self.is_iso = ((endpoint_info.getAttributes() & libusb1.LIBUSB_TRANSFER_TYPE_MASK)
                       == libusb1.LIBUSB_TRANSFER_TYPE_ISOCHRONOUS)
        maxpacketsize = endpoint_info.getMaxPacketSize()
        self.packetsize = ((maxpacketsize >> 11)+1) * (maxpacketsize & 0x7ff)
        return True


    def read_async_iso(self, callback, packets, outstanding_transfers, raw):
        """
        Internal function to read from isochronous channel.  External
        users should call read_async.
        """
        array_builder = array.array
        shutdown_event = threading.Event()
        shutdown_is_set = shutdown_event.is_set
        if self.num_channels == 1 and raw:
            def transfer_callback(iso_transfer):
                for (status, data) in iso_transfer.iterISO():
                    callback(data, '')
                if not shutdown_is_set():
                    iso_transfer.submit()
        elif self.num_channels == 1 and not raw:
            def transfer_callback(iso_transfer):
                for (status, data) in iso_transfer.iterISO():
                    callback(array_builder('B', data), [])
                if not shutdown_is_set():
                    iso_transfer.submit()
        elif self.num_channels == 2 and raw:
            def transfer_callback(iso_transfer):
                for (status, data) in iso_transfer.iterISO():
                    callback(data[::2], data[1::2])
                if not shutdown_is_set():
                    iso_transfer.submit()
        elif self.num_channels == 2 and not raw:
            def transfer_callback(iso_transfer):
                for (status, data) in iso_transfer.iterISO():
                    callback(array_builder('B', data[::2]), array_builder('B', data[1::2]))
                if not shutdown_is_set():
                    iso_transfer.submit()
        else:
            assert False
        for _ in range(outstanding_transfers):
            transfer = self.device_handle.getTransfer(iso_packets=packets)
            transfer.setIsochronous(0x82, (packets*self.packetsize), callback=transfer_callback)
            transfer.submit()
        return shutdown_event


    def read_async_bulk(self, callback, packets, outstanding_transfers, raw):
        """
        Internal function to read from bulk channel.  External
        users should call read_async.
        """
        array_builder = array.array
        shutdown_event = threading.Event()
        shutdown_is_set = shutdown_event.is_set
        if self.num_channels == 1 and raw:
            def transfer_callback(bulk_transfer):
                data = bulk_transfer.getBuffer()[0:bulk_transfer.getActualLength()]
                callback(data, '')
                if not shutdown_is_set():
                    bulk_transfer.submit()
        elif self.num_channels == 1 and not raw:
            def transfer_callback(bulk_transfer):
                data = bulk_transfer.getBuffer()[0:bulk_transfer.getActualLength()]
                callback(array_builder('B', data), [])
                if not shutdown_is_set():
                    bulk_transfer.submit()
        elif self.num_channels == 2 and raw:
            def transfer_callback(bulk_transfer):
                data = bulk_transfer.getBuffer()[0:bulk_transfer.getActualLength()]
                callback(data[::2], data[1::2])
                if not shutdown_is_set():
                    bulk_transfer.submit()
        elif self.num_channels == 2 and not raw:
            def transfer_callback(bulk_transfer):
                data = bulk_transfer.getBuffer()[0:bulk_transfer.getActualLength()]
                callback(array_builder('B', data[::2]), array_builder('B', data[1::2]))
                if not shutdown_is_set():
                    bulk_transfer.submit()
        else:
            assert False
        for _ in range(outstanding_transfers):
            transfer = self.device_handle.getTransfer(iso_packets=packets)
            transfer.setBulk(0x86, (packets*self.packetsize), callback=transfer_callback)
            transfer.submit()
        return shutdown_event


    def read_async(self, callback, data_size, outstanding_transfers=3, raw=False):
        """
        Read both channel's ADC data from the device asynchronously. No trigger support, you need to do this in software.
        The function returns immediately but the data is then sent asynchronously to the callback function whenever it
        receives new samples.
        :param callback: A function with two arguments that take the samples for the first and second channel.
        :param data_size: The block size for each sample.  This is automatically rounded up to the nearest multiple of
                          the native block size.
        :param int outstanding_transfers: (OPTIONAL) The number of transfers sent to the kernel at the same time to
                improve gapless sampling.  The higher, the more likely it works, but the more resources it will take.
        :param raw: (OPTIONAL) Whether the samples should be returned as raw string (8-bit data) or as an array of bytes.
        :return: Returns a shutdown event handle if successful (and then calls the callback asynchronously).
                 Call set() on the returned event to stop sampling.
        """
        # data_size to packets
        packets = (data_size + self.packetsize-1)//self.packetsize
        if self.is_iso:
            return self.read_async_iso(callback, packets, outstanding_transfers, raw)
        else:
            return self.read_async_bulk(callback, packets, outstanding_transfers, raw)


    def scale_read_data( self, read_data, voltage_range=1, channel=1, probe=1, offset=0 ):
        """
        Convenience function for converting data read from the scope to nicely scaled voltages.
        Apply the calibration values that are stored in EEPROM (call "get_calibration_values()" before)
        :param list read_data: The list of points returned from the read_data functions.
        :param int voltage_range: The voltage range current set for the channel.
        :param int channel: 1 = CH1, 2 = CH2.
        :param int probe_multiplier: (OPTIONAL) An additonal multiplictive factor for changing the probe gain.
                                 Default: 1
        :param int offset: (OPTIONAL) An additional additive value to compensate the ADC offset
        :return: A list of correctly scaled voltages for the data.
        """
        if channel == 1:
                mul = probe * self.gain1[ voltage_range ]
                off = offset + self.offset1[ voltage_range ]
        else:
                mul = probe * self.gain2[ voltage_range ]
                off = offset + self.offset2[ voltage_range ]

        scale_factor = ( 5.12 * mul ) / ( voltage_range << 7 )
        return [ ( datum - 128 - off ) * scale_factor for datum in read_data ]


    def voltage_to_adc( self, voltage, voltage_range=1, channel=1, probe=1, offset=0 ):
        """
        Convenience function for analog voltages into the ADC count the scope would see.
        :param float voltage: The analog voltage to convert.
        :param int voltage_range: The voltage range current set for the channel.
        :param int channel: 1 = CH1, 2 = CH2.
        :param int probe: (OPTIONAL) An additonal multiplictive factor for changing the probe gain.
                                 Default: 1
        :param int offset: (OPTIONAL) An additional additive value to simulate the ADC offset
        :return: The corresponding ADC count.
        """
        if channel == 1:
                mul = probe * self.gain1[ voltage_range ]
                off = offset + self.offset1[ voltage_range ]
        else:
                mul = probe * self.gain2[ voltage_range ]
                off = offset + self.offset2[ voltage_range ]
        return voltage * ( voltage_range << 7 ) / ( 5.12 * mul ) + 128 + off


    def adc_to_voltage( self, adc_count, voltage_range=1, channel=1, probe=1, offset=0 ):
        """
        Convenience function for converting an ADC count from the scope to a nicely scaled voltage.
        :param int adc_count: The scope ADC count.
        :param int voltage_range: The voltage range current set for the channel.
        :param int channel: 1 = CH1, 2 = CH2.
        :param int probe: (OPTIONAL) An additonal multiplictive factor for changing the probe gain.
                                 Default: 1
        :param int offset: (OPTIONAL) An additional additive value to correct the ADC offset, default: 0
        :return: The analog voltage corresponding to that ADC count.
        """
        if channel == 1:
                mul = probe * self.gain1[ voltage_range ]
                off = offset + self.offset1[ voltage_range ]
        else:
                mul = probe * self.gain2[ voltage_range ]
                off = offset + self.offset2[ voltage_range ]
        return ( adc_count - 128 - off ) * ( 5.12 * mul ) / ( voltage_range << 7 )


    def set_sample_rate(self, rate_index, timeout=0):
        """
        Set the sample rate index for the scope to sample at. This determines the time between each point the scope
        returns.
        :param rate_index: The rate_index. These are the keys for the SAMPLE_RATES dict for the Oscilloscope object.
                           Common rate_index values and actual sample rate per channel:
                           102 <->  20 kS/s
                           104 <->  40 kS/s
                           105 <->  50 kS/s
                           106 <->  64 kS/s
                           110 <-> 100 kS/s
                           120 <-> 200 kS/s
                           140 <-> 400 kS/s
                           150 <-> 500 kS/s
                             1 <->   1 MS/s
                             2 <->   2 MS/s
                             3 <->   3 MS/s
                             4 <->   4 MS/s
                             5 <->   5 MS/s
                             6 <->   6 MS/s
                             8 <->   8 MS/s
                            10 <->  10 MS/s
                            12 <->  12 MS/s
                            15 <->  15 MS/s
                            16 <->  16 MS/s
                            24 <->  24 MS/s
                            30 <->  30 MS/s
                            48 <->  48 MS/s

                           Other values are not supported.
        :param timeout:
        :return: True if successful. This method may assert or raise various libusb errors if something went wrong.
        """
        if not self.device_handle:
            assert self.open_handle()
        bytes_written = self.device_handle.controlWrite(0x40, self.SET_SAMPLE_RATE_REQUEST,
                                                        self.SET_SAMPLE_RATE_VALUE,
                                                        self.SET_SAMPLE_RATE_INDEX,
                                                        pack("B", rate_index), timeout=timeout)
        assert bytes_written == 0x01
        return True


    def convert_sampling_rate_to_measurement_times(self, num_points, rate_index):
        """
        Convenience method for converting a sampling rate index into a list of times from beginning of data collection
        and getting human-readable sampling rate string.
        :param num_points: The number of data points.
        :param rate_index: The sampling rate index used for data collection.
        :return: A list of times in seconds from beginning of data collection, and the nice human readable rate label.
        """
        rate_label, rate = self.SAMPLE_RATES.get(rate_index, ("? MS/s", 1.0))
        return [i/rate for i in range(num_points)], rate_label


    def set_num_channels(self, nchannels, timeout=0):
        """
        Set the number of active channels.  Either we sample only CH1 or we
        sample CH1 and CH2.
        :param nchannels: The number of active channels.  This is 1 or 2.
        :param timeout: (OPTIONAL).
        :return: True if successful. This method may assert or raise various libusb errors if something went wrong.
        """
        if not self.supports_single_channel:
            return False
        assert nchannels == 1 or nchannels == 2
        if not self.device_handle:
            assert self.open_handle()
        bytes_written = self.device_handle.controlWrite(0x40, self.SET_NUMCH_REQUEST,
                                                        self.SET_NUMCH_VALUE,
                                                        self.SET_NUMCH_INDEX,
                                                        pack("B", nchannels), timeout=timeout)
        assert bytes_written == 0x01
        self.num_channels = nchannels
        return True


    def set_ch1_voltage_range(self, range_index, timeout=0):
        """
        Set the voltage scaling factor at the scope for channel 1 (CH1).
        :param range_index: A numerical constant, which determines the devices range by the following formula:
                            range := +/- 5.0 V / (range_indeX).

                            The stock software only typically uses the range indicies 0x01, 0x02, 0x05, and 0x0a,
                            but others, such as 0x08 and 0x0b seem to work correctly.

                            This same range_index is given to the scale_read_data method to get nicely scaled
                            data in voltages returned from the scope.

        :param timeout: (OPTIONAL).
        :return: True if successful. This method may assert or raise various libusb errors if something went wrong.
        """
        if not self.device_handle:
            assert self.open_handle()
        bytes_written = self.device_handle.controlWrite(0x40, self.SET_CH1_VR_REQUEST,
                                                        self.SET_CH1_VR_VALUE,
                                                        self.SET_CH1_VR_INDEX,
                                                        pack("B", range_index), timeout=timeout)
        assert bytes_written == 0x01
        return True


    def set_ch2_voltage_range(self, range_index, timeout=0):
        """
        Set the voltage scaling factor at the scope for channel 1 (CH1).
        :param range_index: A numerical constant, which determines the devices range by the following formula:
                            range := +/- 5.0 V / (range_indeX).

                            The stock software only typically uses the range indicies 0x01, 0x02, 0x05, and 0x0a,
                            but others, such as 0x08 and 0x0b seem to work correctly.

                            This same range_index is given to the scale_read_data method to get nicely scaled
                            data in voltages returned from the scope.

        :param timeout: (OPTIONAL).
        :return: True if successful. This method may assert or raise various libusb errors if something went wrong.
        """
        if not self.device_handle:
            assert self.open_handle()
        bytes_written = self.device_handle.controlWrite(0x40, self.SET_CH2_VR_REQUEST,
                                                        self.SET_CH2_VR_VALUE,
                                                        self.SET_CH2_VR_INDEX,
                                                        pack("B", range_index), timeout=timeout)
        assert bytes_written == 0x01
        return True


    def set_calibration_frequency(self, cal_freq, timeout=0):
        """
        Set the frequency of the calibration output.
        :param cal_freq: 32 Hz <= cal_freq <= 100 kHz
        The frequency will be coded into one byte for the Oscilloscope object.
          0 -> 100 Hz (be compatible with sigrok FW)
          1..100 -> 1..100 kHz
          101..102 -> do not use
          103 -> 32 Hz (lowest possible calfreq due to 16bit HW counter)
          104..200 -> 40..1000 Hz, step = 10 Hz ( calfreq = 10*(freq-100) )
          201..255 -> 100..5500 Hz, step = 100 Hz ( calfreq = 100*(freq-200) )
        :param timeout: (OPTIONAL).
        :return: True if successful. This method may assert or raise various libusb errors if something went wrong.
        """
        if cal_freq < 32 or cal_freq > 100000:
            return False

        if cal_freq < 1000:
            cal_freq_byte = int(cal_freq / 10) + 100 # 103...199 -> 32...990 Hz
        elif cal_freq < 5600:
            cal_freq_byte = int(cal_freq / 100) + 200 # 201...255 -> 100...5500 Hz
        else:
            cal_freq_byte = ( cal_freq / 1000 ) # 1...100 -> 1...100 kHz

        if not self.device_handle:
            assert self.open_handle()

        bytes_written = self.device_handle.controlWrite(0x40, self.SET_CAL_FREQ_REQUEST,
                                                        self.SET_CAL_FREQ_VALUE,
                                                        self.SET_CAL_FREQ_INDEX,
                                                        pack("B", cal_freq_byte), timeout=timeout)
        assert bytes_written == 0x01
        return True


    def set_ch1_ch2_ac_dc(self, ac_dc, timeout=0):
        """
        set AC_DC status of both channels
        :param ac_dc: the value to send:
        0x00: CH2 AC, CH1 AC
        0x01: CH2 AC, CH1 DC
        0x10: CH2 DC, CH1 AC
        0x11: CH2 DC, CH1 DC
        :param timeout: (OPTIONAL).
        :return: True if successful. This method may assert or raise various libusb errors if something went wrong.
        """
        assert ac_dc == self.AC_AC or ac_dc == self.AC_DC or ac_dc == self.DC_AC or ac_dc == self.DC_DC
        if not self.device_handle:
            assert self.open_handle()
        bytes_written = self.device_handle.controlWrite(0x40, self.SET_AC_DC_REQUEST,
                                                        self.SET_AC_DC_VALUE,
                                                        self.SET_AC_DC_INDEX,
                                                        pack("B", ac_dc), timeout=timeout)
        assert bytes_written == 0x01
        self.ac_dc_status = ac_dc
        return True


    def set_ch1_ac_dc(self, ac_dc, timeout=0):
        """
        set AC_DC status of channel 1
        :param ac_dc: the value to set:
        0x00: AC
        0x01: DC
        :param timeout: (OPTIONAL).
        :return: True if successful. This method may assert or raise various libusb errors if something went wrong.
        """
        assert ac_dc == self.AC or ac_dc == self.DC
        if not self.device_handle:
            assert self.open_handle()
        ac_dc_new = ( self.ac_dc_status & 0xF0 ) | ac_dc
        bytes_written = self.device_handle.controlWrite(0x40, self.SET_AC_DC_REQUEST,
                                                        self.SET_AC_DC_VALUE,
                                                        self.SET_AC_DC_INDEX,
                                                        pack("B", ac_dc_new), timeout=timeout)
        assert bytes_written == 0x01
        self.ac_dc_status = ac_dc_new # remember the latest status
        return True


    def set_ch2_ac_dc(self, ac_dc, timeout=0):
        """
        set AC_DC status of channel 2
        :param ac_dc: the value to set:
        0x00: AC
        0x01: DC
        :param timeout: (OPTIONAL).
        :return: True if successful. This method may assert or raise various libusb errors if something went wrong.
        """
        assert ac_dc == self.AC or ac_dc == self.DC
        if not self.device_handle:
            assert self.open_handle()
        ac_dc_new = ( self.ac_dc_status & 0x0F ) | ac_dc << 4
        bytes_written = self.device_handle.controlWrite(0x40, self.SET_AC_DC_REQUEST,
                                                        self.SET_AC_DC_VALUE,
                                                        self.SET_AC_DC_INDEX,
                                                        pack("B", ac_dc_new), timeout=timeout)
        assert bytes_written == 0x01
        self.ac_dc_status = ac_dc_new # remember the latest status
        return True



import os
import sys
import time

import numpy
import numpy as np
from cffi import FFI


class C_Code:
    """
    This class contains all wrapper functions to call C code.

    Attributes:
        c_code_loaded (bool): Indicates whether the C code could be imported. If not, this class cannot be used.
        ffibuilder (FFI): A FFI instance.
    """

    def __init__(self):
        """
        Standard constructor. Here the C code is compiled and imported.
        """
        self.c_code_loaded = False
        self.ffibuilder = FFI()

        # Here the headers of all C-functions must be specified
        self.ffibuilder.cdef("int find_trigger_position(double *data_array, int length, double threshold, int rising_edge);")
        self.ffibuilder.cdef("double* create_timing_data(double *data_array, int number_of_points, int sample_rate);")
        self.ffibuilder.cdef("double* create_pretrigger_timing_data(double *data_array, int number_of_points, int sample_rate);")
        self.ffibuilder.cdef("double* create_voltage_data(double *data_array, int length, double scale_factor, double offset);")

        c_code_path = os.path.abspath(os.path.dirname(__file__))
        sys.path.insert(0, c_code_path)
        global find_trigger_position, create_timing_data, create_pretrigger_timing_data, \
            create_voltage_data

        # if the C code is compiled, import it. Otherwise, compile and import it.
        try:
            from _triggering.lib import find_trigger_position, create_timing_data, create_pretrigger_timing_data, \
                create_voltage_data
            self.c_code_loaded = True
        except ModuleNotFoundError:
            print("building C code")
            self.ffibuilder.set_source("_triggering",  # name of the output C extension
                                       """ #include "triggering.h" """,
                                       sources=['triggering.c'],  # includes pi.c as additional sources
                                       libraries=[], )  # on Unix, link with the math library
            self.ffibuilder.compile(c_code_path, verbose=False)
            from _triggering.lib import find_trigger_position, create_timing_data, create_pretrigger_timing_data, \
                create_voltage_data
            self.c_code_loaded = True

    def _create_c_array(self, data_array, dtype=float):
        if dtype == float:
            if not isinstance(data_array, numpy.ndarray):
                data_array = np.array(data_array)
            data_array = data_array.astype(float, order='C')
            c_array = self.ffibuilder.from_buffer('double *', data_array)
            return c_array

    def _create_np_array_from_c_array(self, c_array, length):
        return np.frombuffer(self.ffibuilder.buffer(c_array, length * 8), dtype=float)

    def find_trigger_position(self, data_array, threshold, trigger_kind='RISING'):
        """
        Get the array position at which a threshold value is exceeded.

        Args:
            data_array (numpy.array or list): The array containing the measurement data.
            threshold (float): The threshold value at which the trigger should fire.
            trigger_kind (str): The type of trigger (currently 'RISING' and 'FALLING' are supported).

        Returns:
            The array position where the threshold is hit or crossed.

        """
        if not self.c_code_loaded:
            raise ImportError('Could not load C code.')
        else:
            c_array = self._create_c_array(data_array)
            if trigger_kind == 'RISING':
                return find_trigger_position(c_array, len(data_array), threshold, 1)
            else:
                return find_trigger_position(c_array, len(data_array), threshold, 0)

    def create_voltage_data(self, raw_data, scale_factor, offset):
        """
        Convenience function for converting data read from the scope to nicely scaled voltages.

        Args:
            raw_data  (numpy.array or list): The list of points returned from the read_data functions.
            scale_factor (float): A calculated scale factor.
            offset (float): A calculated offset.

        Returns:
            numpy.array: The voltage data for the given raw data.
        """
        if not self.c_code_loaded:
            raise ImportError('Could not load C code.')
        else:
            c_array = self._create_c_array(raw_data)
            create_voltage_data(c_array, len(raw_data), scale_factor, offset)
            np_array = self._create_np_array_from_c_array(c_array, len(raw_data))
            return np_array

    def create_timing_data(self, num_points, sample_rate):
        """
        Convenience method for creating a list of times from the read data.

        Args:
            num_points (int): The number of measured points.
            sample_rate (int): the sample rate of the oscilloscope.

        Returns:
            numpy.array: The timing data.
        """
        if not self.c_code_loaded:
            raise ImportError('Could not load C code.')
        else:
            raw_data = np.zeros(num_points)
            c_array = self._create_c_array(raw_data)
            create_timing_data(c_array, len(raw_data), sample_rate)
            np_array = self._create_np_array_from_c_array(c_array, len(raw_data))
            return np_array

    def create_pretrigger_timing_data(self, num_points, sample_rate):
        """
        Convenience method for creating a list of pre-trigger times from the read data.

        Args:
            num_points (int): The number of measured points.
            sample_rate (int): the sample rate of the oscilloscope.

        Returns:
            numpy.array: The timing data.
        """
        if not self.c_code_loaded:
            raise ImportError('Could not load C code.')
        else:
            raw_data = np.zeros(num_points)
            c_array = self._create_c_array(raw_data)
            create_pretrigger_timing_data(c_array, len(raw_data), sample_rate)
            np_array = self._create_np_array_from_c_array(c_array, len(raw_data))
            return np_array


import hantekosc.py_ht6022.LibUsbScope


def list_connected_hantek_devices():
    """
    List all connected oscilloscopes from hantek.

    Returns:
        list(dict): A list containing a dict containing 'Manufacturer', 'Model' and 'Serial Number' for each device
    """
    scope = hantekosc.py_ht6022.LibUsbScope.Oscilloscope()
    usb_device_list = scope.context.getDeviceList(skip_on_error=True)
    device_list = []

    for device in usb_device_list:
        # Hantek 6022BE
        if (device.getVendorID() == scope.FIRMWARE_PRESENT_VENDOR_ID or device.getVendorID() ==
                scope.NO_FIRMWARE_VENDOR_ID) and device.getProductID() == scope.PRODUCT_ID_BE:
            hantek_6022be = hantekosc.py_ht6022.LibUsbScope.Oscilloscope(device.getVendorID(), device.getProductID())
            hantek_6022be.setup()
            if not  hantek_6022be.open_handle():
                sys.exit(-1)

            if (not  hantek_6022be.is_device_firmware_present):
                print('upload firmware...')
                hantek_6022be.flash_firmware()
            resource_info = {'Manufacturer': 'Hantek', 'Model': '6022BE',
                             'Serial Number': hantek_6022be.get_serial_number_string()}
            device_list.append(resource_info)
            device.close()
        # Hantek 6022BL
        if (device.getVendorID() == scope.FIRMWARE_PRESENT_VENDOR_ID or device.getVendorID() ==
                scope.NO_FIRMWARE_VENDOR_ID) and device.getProductID() == scope.PRODUCT_ID_BL:
            hantek_6022be = hantekosc.PyHT6022.LibUsbScope.Oscilloscope(device.getVendorID(), device.getProductID())
            hantek_6022be.setup()
            if not hantek_6022be.open_handle():
                sys.exit(-1)

            if (not hantek_6022be.is_device_firmware_present):
                print('upload firmware...')
                hantek_6022be.flash_firmware()
            resource_info = {'Manufacturer': 'Hantek', 'Model': '6022BL',
                             'Serial Number': hantek_6022be.get_serial_number_string()}
            device_list.append(resource_info)
            device.close()
        # Hantek 6021
        if (device.getVendorID() == scope.FIRMWARE_PRESENT_VENDOR_ID or device.getVendorID() ==
                scope.NO_FIRMWARE_VENDOR_ID) and device.getProductID() == scope.PRODUCT_ID_21:
            hantek_6022be = hantekosc.PyHT6022.LibUsbScope.Oscilloscope(device.getVendorID(), device.getProductID())
            hantek_6022be.setup()
            if not hantek_6022be.open_handle():
                sys.exit(-1)

            if (not hantek_6022be.is_device_firmware_present):
                print('upload firmware...')
                hantek_6022be.flash_firmware()
            resource_info = {'Manufacturer': 'Hantek', 'Model': '6021',
                             'Serial Number': hantek_6022be.get_serial_number_string()}
            device_list.append(resource_info)
            device.close()
    return device_list



import threading
import numpy as np


class Channel:
    """
    Class for an oscilloscope channel.

    Attributes:
        id (str): The channel id.
        ch_number (int): The channel number. Starts with 0.
        osc: The device that owns the channel.


    """
    def __init__(self, osc, channel_number):
        """
        Class constructor.

        Args:
            osc: The device that owns the channel.
            channel_number: The number of the channel.
        """
        self._trigger_kind = 'RISING'
        self._trigger_level = 1

        self.data_mutex = threading.Lock()

        self.voltage_index = 1
        self.id = 'CH' + str(channel_number + 1)
        self.ch_number = channel_number
        self.osc = osc
        self.retrieved_data = np.empty((2, 1), float)
        self.new_data_ready = False
        self.voltage_range = 5

        # ToDo: Add "enabled" variable to improve performance by calculating data only for enabled channels
        # ToDo: Add "probe gain" and "probe offset" variables

    @property
    def measured_data(self):
        """
        Get the measurement data of the x and y coordinates in seconds and volts.

        Returns:
            numpy.array: The measurement data of the x and y coordinates in seconds and volts.
        """
        self.data_mutex.acquire()
        data = np.array(self.retrieved_data, copy=True)
        self.new_data_ready = False
        self.data_mutex.release()
        return data

    @property
    def voltage_ranges_available(self):
        """
        Get available voltage ranges

        Returns:
            list: The available voltage ranges as list.
        """
        return [0.5, 1, 2.5, 5]

    @property
    def voltage_range(self):
        """
        Get the voltage range.

        Returns:
            int: The voltage range.
        """
        return self._voltage_range

    @voltage_range.setter
    def voltage_range(self, voltage_range):
        """
        Set the voltage range.

        Args:
            voltage_range (int): The voltage range.
        """
        self.osc.settings_mutex.acquire()

        was_running = False
        if self.osc.running:
            self.osc.stop()
            was_running = True

        # ToDo: raise value modified error?
        match voltage_range:
            case num if num < 0.75:
                index = 10
            case num if 0.75 <= num < 1.75:
                index = 5
            case num if 1.75 <= num < 3.75:
                index = 2
            case num if 3.75 <= num:
                index = 1
            case _:
                index = 1

        if self.id == 'CH1':
            pass
            self.osc.scope.set_ch1_voltage_range(index)
        elif self.id == 'CH2':
            pass
            self.osc.scope.set_ch2_voltage_range(index)
        self._voltage_range = 5/index
        self.voltage_index = index

        # device must be started and stopped for some reason (Otherwise an SIGSEGV error is thrown)
        if was_running:
            self.osc.start()

        self.osc.settings_mutex.release()

    @property
    def trigger_kinds_available(self):
        """
        Get available trigger kinds

        Returns:
            list: The available trigger kinds as list.
        """
        return ['RISING', 'FALLING']

    @property
    def trigger_kind(self):
        """
        Get the trigger kind ('RISING', 'FALLING').

        Returns:
            str: The trigger kind.

        """
        return self._trigger_kind

    @trigger_kind.setter
    def trigger_kind(self, trigger_kind):
        """
        Set the trigger kind ('RISING', 'FALLING').

        Args:
            trigger_kind (str):  The trigger kind.
        """
        self.osc.settings_mutex.acquire()
        self._trigger_kind = trigger_kind
        self.osc.settings_mutex.release()

    @property
    def trigger_level(self):
        """
        Get the trigger level.

        Returns:
            float: The trigger level.
        """
        return self._trigger_level

    @trigger_level.setter
    def trigger_level(self, trigger_level):
        """
        Set the trigger level
        Args:
            trigger_level (float): The trigger level.
        """
        self.osc.settings_mutex.acquire()
        self._trigger_level = trigger_level
        self.osc.settings_mutex.release()



import os
import sys

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# Add source code directory to path (required for autodoc)
project_path = os.path.join('..', '..', '..')
sys.path.insert(0, os.path.abspath(project_path))

extensions = ['sphinx.ext.napoleon', 'sphinx.ext.autodoc']

templates_path = ['_templates']
exclude_patterns = []


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'hantekosc'
copyright = '2023, Bruno Mecke'
author = 'Bruno Mecke'
release = '0.1.0'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



import threading
import time
import numpy as np
from hantekosc.py_ht6022 import LibUsbScope

from threading import Thread
from queue import Queue

from hantekosc.c_code import C_Code

from hantekosc.channel import Channel


class Oscilloscope:
    """
    Interface for hantek usb oscilloscopes.

    Attributes:
        c_code (C_Code): This object is used to call functions programmed in C language.
        scope (Oscilloscope): A oscilloscope object from Hantek6022API (https://github.com/Ho-Ro/Hantek6022API).
        running (bool): Indicates whether the device is currently running.
        channels (list): A list containing objects for each channel of the device.
        settings_mutex (threading.lock): A mutex ensuring that only one setting can be made at a time.
    """

    def __init__(self, serial_number=None):
        """
        Class constructor. Open the connection to the instrument using the Hantek6022API
        (https://github.com/Ho-Ro/Hantek6022API).
        """
        # ToDo: initialize device with serial number
        self.running = False

        self.c_code = C_Code()

        self.scope = LibUsbScope.Oscilloscope()
        # connect to device
        if not self.scope.setup() or not self.scope.open_handle():
            raise RuntimeError("Could not find any hantek devices")

        # upload correct firmware into device's RAM
        if not self.scope.is_device_firmware_present:
            self.scope.flash_firmware() #firmware=PyHT6022.Firmware.mod_firmware_01)
        self.scope.set_num_channels(2)

        self._blockslope = 0.00000207

        # event to stop async data reading thread
        self._shutdown_event = None
        # calculated when setting presample ratio
        self._number_of_presample_points = 0
        # Size of a block read from the device via USB (calculated when setting sample rate)
        self._blocksize = 0
        self._record_length = 0
        self._sample_rate = 0
        self._sample_id = 0
        self._pre_sample_ratio = 0
        self._trigger_mode = ''
        self._selected_channel = 0

        self._raw_data = Queue(maxsize=50)
        self._previous_voltage_data = [np.zeros(self._record_length), np.zeros(self._record_length)]

        self.settings_mutex = threading.Lock()

        # initial oscilloscope settings
        self.sample_rate = 20 * 1e3
        self.record_length = 5000
        self.pre_sample_ratio = 0.5
        self.trigger_mode = 'REPEAT'  # SINGLE, AUTO, REPEAT
        self.selected_channel = 0
        self.channels = [Channel(self, 0), Channel(self, 1)]

    def __del__(self):
        """
        Class destructor. Stop the device when the program is closed.
        """
        if self.running:
            self.stop()
            self.scope.close_handle()

    def retrieve_callback(self, ch1_data, ch2_data):
        """
        This callback is called whenever new measurement data is available.
        The data is then putted into a queue to be processed in another thread.

        Args:
            ch1_data (list): Measurement data of the first channel.
            ch2_data (list): Measurement data of the second channel.
        """
        # append data to queue
        if len(ch1_data) == self._blocksize and len(ch2_data) == self._blocksize:
            self._raw_data.put((ch1_data, ch2_data))

    def retrieve(self):
        """
        Here data is permanently requested in a separate thread as long as the device is running.
        """
        while self.running:
            self.scope.poll()

    def start(self):
        """
        Start the measurement.
        """
        self._previous_voltage_data = [np.zeros(self._record_length), np.zeros(self._record_length)]
        self.running = True

        process_data_thread = Thread(target=self._process_data)
        process_data_thread.start()

        self.scope.start_capture()
        self._shutdown_event = self.scope.read_async(self.retrieve_callback, 2 * self._blocksize,
                                                     outstanding_transfers=10)

        retriever_thread = Thread(target=self.retrieve)
        retriever_thread.start()

    def stop(self):
        """
        Stop the measurement.
        """
        self.running = False
        self._shutdown_event.set()
        time.sleep(1)
        self._raw_data.queue.clear()

    def _process_data(self):
        """
        Here the measurement data are processed in a separate thread.
        An array with the time units and an array with the measurement data starting from the trigger point are created.
        In addition, an array is created in which the measurement data is stored that is temporally located before the
        trigger.
        This function is time critical. If the function takes too long, the queue fills up and measurement data is lost.
        """
        while self.running:
            # Although this function is time-critical, a short time can be waited to avoid overloading the processor.
            # This time might depend on the hardware of the user device.
            time.sleep(0.01)
            self.settings_mutex.acquire()
            loop_is_to_slow = False
            if self.running:
                voltage_data = self._get_voltage_data_block()
                if self.trigger_mode == 'SINGLE' or self.trigger_mode == 'REPEAT':
                    trigger_position = self._find_trigger_position(voltage_data)
                    end_position = self.record_length - self._number_of_presample_points
                    # system triggered
                    if trigger_position >= 0:
                        # add points before trigger event to _previous_raw_data array
                        for i in range(self.number_of_channels):
                            prev_data = self._previous_voltage_data[i]
                            prev_data = np.hstack((prev_data, voltage_data[i][:trigger_position]))
                            self._previous_voltage_data[i] = prev_data[len(prev_data) - self.record_length:]
                            voltage_data[i] = voltage_data[i][trigger_position:]
                        # More data is needed than available in one block --> More blocks must be loaded
                        if trigger_position + end_position > self._blocksize:
                            # Points that do not fit into the first block
                            remaining_points = ((self.record_length - self._number_of_presample_points) -
                                                (self._blocksize - trigger_position))
                            # Number of blocks that still need to be loaded
                            number_of_blocks = int(np.ceil(remaining_points / self._blocksize))
                            # get needed blocks
                            for i in range(number_of_blocks):
                                voltage_data_block = self._get_voltage_data_block()
                                for j in range(self.number_of_channels):
                                    voltage_data[j] = np.hstack((voltage_data[j], voltage_data_block[j]))

                    # Calculate the array with the timing points
                    presample_timing_data = self._create_presample_timing_data(self._number_of_presample_points)
                    timing_data_ = self._create_timing_data(self.record_length - self._number_of_presample_points)
                    timing_data = np.hstack((presample_timing_data, timing_data_))

                    # Merge the array containing the measurement data from pre-trigger data
                    # and the data starting from the trigger event
                    for i in range(self.number_of_channels):
                        previous_data_start_pos = len(self._previous_voltage_data[i]) - self._number_of_presample_points
                        voltage_data[i] = np.hstack(
                            (self._previous_voltage_data[i][previous_data_start_pos:], voltage_data[i]))
                        if trigger_position >= 0:
                            self.channels[i].retrieved_data = np.vstack(
                                (timing_data, voltage_data[i][:self.record_length]))
                        self._previous_voltage_data[i] = voltage_data[i]
                        self._previous_voltage_data[i] = self._previous_voltage_data[i][
                                                         len(self._previous_voltage_data[i]) - self.record_length:]
                        self.channels[i].new_data_ready = True
                    # If the trigger mode is "SINGLE", stop after a trigger event
                    if self.trigger_mode == 'SINGLE' and trigger_position >= 0:
                        self.stop()
                # No trigger, just get data blocks
                else:
                    timing_data = self._create_timing_data(self.record_length)
                    # More data is needed than available in one block --> More blocks must be loaded
                    if self.record_length > self._blocksize:
                        number_of_blocks = int(np.ceil((self.record_length - self._blocksize) / self._blocksize))
                        for i in range(number_of_blocks):
                            voltage_data_block = self._get_voltage_data_block()
                            for j in range(self.number_of_channels):
                                voltage_data[j] = np.hstack((voltage_data[j], voltage_data_block[j]))

                    for i in range(self.number_of_channels):
                        self.channels[i].data_mutex.acquire()
                        self.channels[i].retrieved_data = np.vstack((timing_data, voltage_data[i][:self.record_length]))
                        self.channels[i].new_data_ready = True
                        self.channels[i].data_mutex.release()

                # clear queue if the program is too slow
                if self._raw_data.qsize() > 48:
                    print('Data processing is too slow.')
                    for i in range(48):
                        self._raw_data.get()
                    loop_is_to_slow = True

            self.settings_mutex.release()
            if loop_is_to_slow:
                self.pre_sample_ratio = 0

    def _find_trigger_position(self, voltage_data):
        """
        Get the array position at which the trigger level value is exceeded.

        Args:
            voltage_data (list): The array containing the measurement data for both channels.
        Returns:
            The array position of the selected channel where the trigger level value is hit or crossed.
        """

        # find rising or falling edge
        # https://stackoverflow.com/questions/50365310/python-rising-falling-edge-oscilloscope-like-trigger
        selected_channel = self.channels[self.selected_channel]

        if self.c_code.c_code_loaded:
            trigger_pos = self.c_code.find_trigger_position(voltage_data[self.selected_channel],
                                                            selected_channel.trigger_level,
                                                            selected_channel.trigger_kind)
            return trigger_pos
        else:
            # find rising or falling edge
            # https://stackoverflow.com/questions/50365310/python-rising-falling-edge-oscilloscope-like-trigger
            thresholded_array = np.where(voltage_data[self.selected_channel] > selected_channel.trigger_level, 1, -1)
            convolution_array = [1, -1]
            convolution = np.convolve(thresholded_array, convolution_array, mode='same')
            if self.channels[self.selected_channel].trigger_kind == 'RISING':
                argmax = np.argmax(convolution)
                if convolution[argmax] == len(convolution_array):
                    trigger_pos = argmax
                else:
                    trigger_pos = -1
            else:
                argmin = np.argmin(convolution)
                if convolution[argmin] == -1 * len(convolution_array):
                    trigger_pos = argmin
                else:
                    trigger_pos = -1
            return trigger_pos

    def _get_voltage_data_block(self):
        """
        Here a data block is fetched from the queue and the measurement data is converted into volts.

        Returns:
            list: The measurement data in volts.
        """
        raw_data = self._raw_data.get()
        voltage_data = []
        for i in range(self.number_of_channels):
            voltage_data.append(np.array(self._create_voltage_data(raw_data[i], self.channels[i].voltage_index,
                                self.selected_channel)))
        return voltage_data

    def _create_voltage_data(self, raw_data, voltage_range=1, channel=0, probe=1, offset=0):
        """
        Convenience function for converting data read from the scope to nicely scaled voltages.
        Apply the calibration values that are stored in EEPROM (call "get_calibration_values()" before)

        Args:
            raw_data (list): The list of points returned from the read_data functions.
            voltage_range (int): The voltage range current set for the channel.
            channel (int): The voltage range current set for the channel. 0 = CH1, 1 = CH2.
            probe (int): An additonal multiplictive factor for changing the probe gain.
            offset (float): An additional additive value to compensate the ADC offset

        Returns:
            numpy.array: A list of correctly scaled voltages for the data.
        """
        if channel == 0:
                mul = probe * self.scope.gain1[voltage_range]
                off = offset + self.scope.offset1[voltage_range]
        else:
                mul = probe * self.scope.gain2[voltage_range]
                off = offset + self.scope.offset2[voltage_range]

        scale_factor = (5.12 * mul) / (voltage_range << 7)

        if self.c_code.c_code_loaded:
            return self.c_code.create_voltage_data(raw_data, scale_factor, off)
        else:
            data = np.array(raw_data, dtype=float)
            data = (data - 128 - off) * scale_factor
            return data

    def _create_timing_data(self, num_points):
        """
        Convenience method for creating a list of times from the read data.

        Args:
            num_points (int): The number of measured points.

        Returns:
            numpy.array: The timing data.
        """
        if self.c_code.c_code_loaded:
            timing_data = self.c_code.create_timing_data(num_points, self.sample_rate)
            return timing_data
        else:
            timing_data, _ = self.scope.convert_sampling_rate_to_measurement_times(num_points, self._sample_id)
            return np.array(timing_data)

    def _create_presample_timing_data(self, num_points):
        """
        Convenience method for creating a list of pre-trigger times from the read data.

        Args:
            num_points (int): The number of measured points.

        Returns:
            numpy.array: The timing data.
        """
        if num_points > 0:
            if self.c_code.c_code_loaded:
                return self.c_code.create_pretrigger_timing_data(num_points, self.sample_rate)
            else:
                timing_data = np.arange(-num_points, 0)
                timing_data = (timing_data / self.sample_rate)
                return timing_data
        else:
            return np.array([])

    @property
    def max_sample_rate(self):
        """
        Get the maximum sample rate in Hz.

        Returns:
            int: The maximum sample rate in Hz.
        """
        return 48 * 1e6

    @property
    def sample_rate(self):
        """
        Get the current sample rate in Hz.

        Returns:
            float: The sample rate in Hz.
        """
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, sample_rate):
        """
        Set the current sample rate in Hz.

        Args:
            sample_rate(int): The sample rate in Hz.
        """
        self.settings_mutex.acquire()
        was_running = False
        if self.running:
            self.stop()
            was_running = True
        # ToDo: rase value modified error?
        match sample_rate:
            case num if num in range(0, 30000):
                sample_id = 102
            case num if num in range(30000, 45000):
                sample_id = 104
            case num if num in range(45000, 57000):
                sample_id = 105
            case num if num in range(57000, 82000):
                sample_id = 106
            case num if num in range(82000, 150000):
                sample_id = 110
            case num if num in range(150000, 300000):
                sample_id = 120
            case num if num in range(300000, 450000):
                sample_id = 140
            case num if num in range(450000, 750000):
                sample_id = 150
            case num if num in range(750000, 1500000):
                sample_id = 1
            case num if num in range(1500000, 2500000):
                sample_id = 2
            case num if num in range(2500000, 3500000):
                sample_id = 3
            case num if num in range(3500000, 4500000):
                sample_id = 4
            case num if num in range(3500000, 4500000):
                sample_id = 4
            case num if num in range(4500000, 5500000):
                sample_id = 5
            case num if num in range(5500000, 6500000):
                sample_id = 6
            case num if num in range(6500000, 9500000):
                sample_id = 8
            case num if num in range(9500000, 11000000):
                sample_id = 10
            case num if num in range(11000000, 13500000):
                sample_id = 12
            case num if num in range(13500000, 15500000):
                sample_id = 15
            case num if num in range(15500000, 20000000):
                sample_id = 16
            case num if num in range(20000000, 27000000):
                sample_id = 24
            case num if num in range(27000000, 39000000):
                sample_id = 30
            case num if num > 39000000:
                sample_id = 48
            case _:
                # ToDo: raise error?
                sample_id = 0
        self.scope.set_sample_rate(sample_id)
        if sample_id >= 100:
            self._sample_rate = int((sample_id - 100) * 10e3)
        else:
            self._sample_rate = int(sample_id * 1000000)
        self._sample_id = sample_id

        self._blocksize = int(self._blockslope * self.sample_rate + 1) * (6 * 1024)  # (should be divisible by 6*1024)

        if was_running:
            self.start()
        self.settings_mutex.release()

    @property
    def number_of_channels(self):
        """
        Get the number of available channels.

        Returns:
            int: The number of channels.
        """
        return len(self.channels)

    @property
    def pre_sample_ratio(self):
        """
        Get the pre sample ratio.

        The pre sample ratio is a float between 0 and 1 and defines how many
        samples should be recorded before the trigger point.

        Returns:
            float: The pre sample ratio.

        """
        return self._pre_sample_ratio

    @pre_sample_ratio.setter
    def pre_sample_ratio(self, ratio):
        """
        Set the pre sample ratio.

        The pre sample ratio is a float between 0 and 1 and defines how many
        samples should be recorded before the trigger point.

        Args:
            ratio (float): The pre sample ratio.

        """
        self.settings_mutex.acquire()
        self._pre_sample_ratio = ratio
        self._number_of_presample_points = int(self.record_length * self._pre_sample_ratio)
        self.settings_mutex.release()

    @property
    def max_record_length(self):
        """
        Get the maximum record length.

        Returns:
            int: The maximum record length.
        """
        return 100000

    @property
    def record_length(self):
        """
        Get the current record length (number of samples).

        Returns:
            int: The current record length.

        """
        return self._record_length

    @record_length.setter
    def record_length(self, record_length):
        """
        Set the current record length (number of samples).

        Args:
            record_length (int): The current record length.
        """
        self.settings_mutex.acquire()
        self._record_length = record_length
        self._number_of_presample_points = int(record_length * self._pre_sample_ratio)
        self._previous_voltage_data = [np.zeros(record_length), np.zeros(record_length)]
        self.settings_mutex.release()

    @property
    def trigger_modes_available(self):
        """
        Get available trigger modes

        Returns:
            list: The available trigger modes as list.
        """
        return ['SINGLE', 'REPEAT', 'NONE']

    @property
    def trigger_mode(self):
        """
        Get the current trigger mode ('SINGLE', 'REPEAT' or 'NONE')

        Returns:
            str: The current trigger mode.
        """
        return self._trigger_mode

    @trigger_mode.setter
    def trigger_mode(self, trigger_mode):
        """
        Set the current trigger mode ('SINGLE', 'REPEAT' or 'NONE')

        Args:
            trigger_mode (str): The current trigger mode.

        """
        self.settings_mutex.acquire()
        self._trigger_mode = trigger_mode
        self.settings_mutex.release()

    @property
    def selected_channel(self):
        """
        Get the number of the selected channel.

        Returns:
            int: The number of the selected channel.
        """
        return self._selected_channel

    @selected_channel.setter
    def selected_channel(self, selected_channel):
        """
        Set the number of the selected channel.

        Args:
            selected_channel (int): The number of the selected channel.

        """
        self.settings_mutex.acquire()
        self._selected_channel = selected_channel
        self.settings_mutex.release()



__author__ = 'Robert Cope'

from collections import namedtuple
import os
import array

# Firmwares to bootstrap the Hantek 6022BE Device are located in this module.
# Format: (Data Len, Value, Data) (Index is always 0x00).
FirmwareControlPacket = namedtuple('FirmwareControlPacket', ['size', 'value', 'data'])


def fx2_ihex_to_control_packets(firmware_location):
    packets = list()
    # disable 8051
    packets.append(FirmwareControlPacket(1, 0xe600, b'\x01'))
    with open(firmware_location, 'r') as f:
        for line in f.readlines():
            line = line.strip()
            assert line.startswith(":")
            record_len = int(line[1:3], 16)
            addr = int(line[3:7], 16)
            record_type = int(line[7:9], 16)
            raw_data = line[9:-2]
            record_data = [int(raw_data[i:i+2], 16) for i in range(0, len(raw_data), 2)]
            file_checksum = int(line[-2:], 16)
            assert record_len == len(record_data)
            if record_type == 0x00:
                checksum = (sum(record_data) + record_len + (addr % 256) + (addr >> 8)) % 256
                assert not ((checksum + file_checksum) % 256) & 0xFF
                packets.append(FirmwareControlPacket(record_len, addr, array.array('B', record_data).tobytes()))
            elif record_type == 0x01:
                assert file_checksum == 0xFF
                break
            else:
                raise ValueError('Unknown record type 0x{:2x} encountered!'.format(record_type))
    # enable 8051
    packets.append(FirmwareControlPacket(1, 0xe600, b'\x00'))
    return packets

base_path = os.path.dirname(os.path.realpath(__file__))
stock_firmware = fx2_ihex_to_control_packets(os.path.join(base_path, 'stock', 'stock_fw.ihex'))
mod_firmware_01 = fx2_ihex_to_control_packets(os.path.join(base_path, 'modded', 'mod_fw_01.ihex'))
mod_firmware_iso = fx2_ihex_to_control_packets(os.path.join(base_path, 'modded', 'mod_fw_iso.ihex'))
dso6021_firmware = fx2_ihex_to_control_packets(os.path.join(base_path, 'DSO6021', 'dso6021-firmware.hex'))
dso6022be_firmware = fx2_ihex_to_control_packets(os.path.join(base_path, 'DSO6022BE', 'dso6022be-firmware.hex'))
dso6022bl_firmware = fx2_ihex_to_control_packets(os.path.join(base_path, 'DSO6022BL', 'dso6022bl-firmware.hex'))
dds120_firmware = fx2_ihex_to_control_packets(os.path.join(base_path, 'DDS120', 'dds120-firmware.hex'))
firmware = default_firmware = dso6022be_firmware



import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

from hantekosc import devices
from hantekosc import oscilloscope

print(devices.list_connected_hantek_devices())

osc = oscilloscope.Oscilloscope()
osc.channels[0].voltage_range = 5
osc.channels[1].voltage_range = 5
osc.trigger_mode = 'REPEAT'
osc.sample_rate = 1 * 1e6
osc.record_length = 100000
osc.pre_sample_ratio = 0.2
osc.channels[0].trigger_kind = 'FALLING'

osc.start()


delay_ = 0
case_ = 0

fig, ax = plt.subplots()
xdata, ydata = [], []
ln, = ax.plot([], [], 'b')




def init():
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    return ln,


def update(frame):
    global delay_
    global case_

    x_data = osc.channels[0].measured_data[0]
    y_data = osc.channels[0].measured_data[1]

    """
    if delay_ == 20:
        match case_:
            case 0:
                osc.sample_rate = 50 * 1e3
            case 1:
                osc.sample_rate = 70 * 1e3
            case 3:
                osc.sample_rate = 100 * 1e3
            case 4:
                osc.sample_rate = 500 * 1e3
            case 5:
                osc.sample_rate = 24000 * 1e3
                case_ = 0

        delay_ = 0
        case_ += 1

    delay_ += 1
    """


    ln.set_data(x_data, y_data)

    x_min = min(x_data)
    x_max = max(x_data)
    x_scale = max([abs(x_min), abs(x_max)])
    y_min = min(y_data)
    y_max = max(y_data)
    y_scale = max([abs(y_min), abs(y_max)])

    ln.axes.set_xlim(min(x_data) - x_scale * 0.05, max(x_data) + x_scale * 0.05)
    ln.axes.set_ylim(min(y_data) - y_scale * 0.1, max(y_data) + y_scale * 0.05)
    return ln,

ani = FuncAnimation(fig, update, frames=10, init_func=init, blit=False)
plt.show()

osc.stop()
osc.scope.close_handle()



